# This file is part of the pyBinSim project.
#
# Copyright (c) 2017 A. Neidhardt, F. Klein, N. Knoop, T. Köllmer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import multiprocessing

import numpy as np
import pyfftw
from past.builtins import xrange

nThreads = multiprocessing.cpu_count()


class ConvolverFFTW(object):
    """
    Class for convolving mono (usually for virtual sources) or stereo input (usually for HP compensation)
    with a BRIRsor HRTF
    """

    def __init__(self, ir_size, block_size, process_stereo):

        self.log = logging.getLogger("pybinsim.ConvolverFFTW")
        self.log.info("Convolver: init")

        # Get Basic infos
        self.IR_size = ir_size
        self.block_size = block_size

        # floor (integer) division in python 2 & 3
        self.IR_blocks = self.IR_size // block_size

        # Calculate crossfade windows
        self.crossFadeIn = np.zeros(self.block_size, dtype='float32')
        self.crossFadeOut = np.zeros(self.block_size, dtype='float32')
        self.crossFadeIn[:] = xrange(0, self.block_size)
        # float division in python 2 & 3
        self.crossFadeIn *= 1 / float((self.block_size - 1))
        self.crossFadeOut[:] = np.flipud(self.crossFadeIn)

        # Create default filter and fftw plan
        # Filter format: [nBlocks,blockSize*4]
        # 0 to blockSize*2: left filter
        # blockSize*2 to blockSize*4: right filter
        self.default_filter = pyfftw.zeros_aligned([self.IR_blocks, 2 * (self.block_size + 1)], np.dtype(np.float32))

        self.filter_fftw_plan = pyfftw.builders.rfft(np.zeros(self.block_size * 2), overwrite_input=True,
                                                     planner_effort='FFTW_MEASURE',
                                                     threads=nThreads)

        # Create Input Buffers and create fftw plans
        self.buffer = pyfftw.zeros_aligned(self.block_size * 2, dtype='float32')
        self.bufferFftPlan = pyfftw.builders.rfft(self.buffer, overwrite_input=True,
                                                  planner_effort='FFTW_MEASURE', threads=nThreads)

        self.buffer2 = pyfftw.zeros_aligned(self.block_size * 2, dtype='float32')
        self.buffer2FftPlan = pyfftw.builders.rfft(self.buffer2, overwrite_input=True,
                                                   planner_effort='FFTW_MEASURE', threads=nThreads)

        # Create arrays for the filters and the FDLs.
        # Note: Probably these do not need to be memory aligned
        self.TF_left_blocked = pyfftw.zeros_aligned((self.IR_blocks, self.block_size + 1), dtype='complex64')
        self.TF_right_blocked = pyfftw.zeros_aligned((self.IR_blocks, self.block_size + 1), dtype='complex64')
        self.TF_left_blocked_previous = pyfftw.zeros_aligned((self.IR_blocks, self.block_size + 1), dtype='complex64')
        self.TF_right_blocked_previous = pyfftw.zeros_aligned((self.IR_blocks, self.block_size + 1), dtype='complex64')

        self.FDL_left = pyfftw.zeros_aligned(self.IR_blocks * (self.block_size + 1), dtype='complex64')
        self.FDL_right = pyfftw.zeros_aligned(self.IR_blocks * (self.block_size + 1), dtype='complex64')

        # Arrays for the result of the complex multiply and add
        # These should be memory aligned because ifft is performed with these data
        self.resultLeftFreq = pyfftw.zeros_aligned(self.block_size + 1, dtype='complex64')
        self.resultRightFreq = pyfftw.zeros_aligned(self.block_size + 1, dtype='complex64')
        self.resultLeftFreqPrevious = pyfftw.zeros_aligned(self.block_size + 1, dtype='complex64')
        self.resultRightFreqPrevious = pyfftw.zeros_aligned(self.block_size + 1, dtype='complex64')
        self.resultLeftIFFTPlan = pyfftw.builders.irfft(np.zeros(self.block_size + 1, dtype='complex64'),
                                                        overwrite_input=False, planner_effort='FFTW_MEASURE',
                                                        threads=nThreads)
        self.resultRightIFFTPlan = pyfftw.builders.irfft(np.zeros(self.block_size + 1, dtype='complex64'),
                                                         overwrite_input=False, planner_effort='FFTW_MEASURE',
                                                         threads=nThreads)
        self.resultLeftPreviousIFFTPlan = pyfftw.builders.irfft(np.zeros(self.block_size + 1, dtype='complex64'),
                                                                overwrite_input=False, planner_effort='FFTW_MEASURE',
                                                                threads=nThreads)
        self.resultRightPreviousIFFTPlan = pyfftw.builders.irfft(np.zeros(self.block_size + 1, dtype='complex64'),
                                                                 overwrite_input=False, planner_effort='FFTW_MEASURE',
                                                                 threads=nThreads)

        # Result of the iffft is stored here
        self.outputLeft = pyfftw.zeros_aligned(self.block_size, dtype='float32')
        self.outputRight = pyfftw.zeros_aligned(self.block_size, dtype='float32')

        # Counts how often process() is called
        self.processCounter = 0

        # Flag for interpolation of output blocks (result of process())
        self.interpolate = False

        # Select mono or stereo processing
        self.processStereo = process_stereo

    def get_counter(self):
        """
        Returns processing counter
        :return: processing counter
        """
        return self.processCounter

    def transform_filter(self, filter):
        """
        Transform filter to freq domain

        :param filter:
        :return: transformed filter
        """
        IR_left = filter[:, 0]
        IR_right = filter[:, 1]

        # Split IRs in blocks
        IR_left_blocked = np.reshape(IR_left, (self.IR_blocks, self.block_size))
        IR_right_blocked = np.reshape(IR_right, (self.IR_blocks, self.block_size))

        # Add zeroes to each block
        IR_left_blocked = np.concatenate((IR_left_blocked, np.zeros([self.IR_blocks, self.block_size])), axis=1)
        IR_right_blocked = np.concatenate((IR_right_blocked, np.zeros([self.IR_blocks, self.block_size])), axis=1)

        TF_left_blocked = np.zeros([self.IR_blocks, self.block_size + 1], np.dtype(np.complex64))
        TF_right_blocked = np.zeros([self.IR_blocks, self.block_size + 1], np.dtype(np.complex64))

        for ir_block_count in range(0, self.IR_blocks):
            TF_left_blocked[ir_block_count] = self.filter_fftw_plan(IR_left_blocked[ir_block_count])
            TF_right_blocked[ir_block_count] = self.filter_fftw_plan(IR_right_blocked[ir_block_count])

        return TF_left_blocked, TF_right_blocked

    def setIR(self, filter, do_interpolation):
        """
        Hand over a new set of filters to the convolver
        and define if you want to perform an interpolation/crossfade

        :param filter:
        :param do_interpolation:
        :return: None
        """
        # Save old filters in case interpolation is needed
        self.TF_left_blocked_previous = self.TF_left_blocked
        self.TF_right_blocked_previous = self.TF_right_blocked

        # apply new filters
        self.TF_left_blocked, self.TF_right_blocked = self.transform_filter(filter)

        # Interpolation means cross fading the output blocks (linear interpolation)
        self.interpolate = do_interpolation

    def process_nothing(self):
        """
        Just for testing
        :return: None
        """
        self.processCounter += 1

    def fill_buffer_mono(self, block):
        """
        Copy mono soundblock to input Buffer;
        Transform to Freq. Domain and store result in FDLs
        :param block: Mono sound block
        :return: None
        """

        if block.size < self.block_size:
            # print('Fill up last block')
            block = np.concatenate((block, np.zeros((1, (self.block_size - block.size)))), 1)

        if self.processCounter == 0:
            # insert first block to buffer
            self.buffer[self.block_size:self.block_size * 2] = block

        else:
            # shift buffer
            self.buffer = np.roll(self.buffer, -self.block_size)
            # insert new block to buffer
            self.buffer[self.block_size:self.block_size * 2] = block
            # shift FDLs
            self.FDL_left = np.roll(self.FDL_left, self.block_size + 1)
            self.FDL_right = np.roll(self.FDL_right, self.block_size + 1)

        # transform buffer into freq domain and copy to FDLs
        self.FDL_left[0:self.block_size + 1] = self.FDL_right[0:self.block_size + 1] = self.bufferFftPlan(
            self.buffer)

    def fill_buffer_stereo(self, block):
        """
        Copy stereo soundblock to input Buffer1 and Buffer2;
        Transform to Freq. Domain and store result in FDLs

        :param block:
        :return: None
        """

        if block.size < self.block_size:
            # print('Fill up last block')
            # print(np.shape(block))
            block = np.concatenate((block, np.zeros(((self.block_size - block.size), 2))), 0)

        if self.processCounter == 0:
            # insert first block to buffer
            self.buffer[self.block_size:self.block_size * 2] = block[:, 0]
            self.buffer2[self.block_size:self.block_size * 2] = block[:, 1]

        else:
            # shift buffer
            self.buffer = np.roll(self.buffer, -self.block_size)
            self.buffer2 = np.roll(self.buffer2, -self.block_size)
            # insert new block to buffer
            self.buffer[self.block_size:self.block_size * 2] = block[:, 0]
            self.buffer2[self.block_size:self.block_size * 2] = block[:, 1]
            # shift FDLs
            self.FDL_left = np.roll(self.FDL_left, self.block_size + 1)
            self.FDL_right = np.roll(self.FDL_right, self.block_size + 1)

        # transform buffer into freq domain and copy to FDLs
        self.FDL_left[0:self.block_size + 1] = self.bufferFftPlan(self.buffer)
        self.FDL_right[0:self.block_size + 1] = self.buffer2FftPlan(self.buffer2)

    def multiply_and_add(self, IR_block_count):
        """
        Multiply Current Filters with data stored in the FDL

        :param IR_block_count:
        :return: None
        """

        if IR_block_count == 0:
            self.resultLeftFreq = np.multiply(self.TF_left_blocked[IR_block_count], self.FDL_left[(
                IR_block_count * (self.block_size + 1)):((IR_block_count + 1) * (self.block_size + 1))])

            self.resultRightFreq = np.multiply(self.TF_right_blocked[IR_block_count], self.FDL_right[(
                IR_block_count * (self.block_size + 1)):((IR_block_count + 1) * (self.block_size + 1))])
        else:
            self.resultLeftFreq += np.multiply(self.TF_left_blocked[IR_block_count],
                                               self.FDL_left[
                                               (IR_block_count * (self.block_size + 1)):(
                                                   (IR_block_count + 1) * (self.block_size + 1))])

            self.resultRightFreq += np.multiply(self.TF_right_blocked[IR_block_count],
                                                self.FDL_right[
                                                (IR_block_count * (self.block_size + 1)):(
                                                    (IR_block_count + 1) * (self.block_size + 1))])

    def multiply_and_add_previous(self, irBlockCount):
        """
        Multiply Previous Filters with data stored in the FDL
        Needed when doing block crossfade/interpolation

        :param irBlockCount:
        :return: None
        """

        if irBlockCount == 0:
            self.resultLeftFreqPrevious = np.multiply(self.TF_left_blocked_previous[irBlockCount], self.FDL_left[(
                irBlockCount * (self.block_size + 1)):((irBlockCount + 1) * (self.block_size + 1))])

            self.resultRightFreqPrevious = np.multiply(self.TF_right_blocked_previous[irBlockCount], self.FDL_right[(
                irBlockCount * (self.block_size + 1)):((irBlockCount + 1) * (self.block_size + 1))])
        else:
            self.resultLeftFreqPrevious += np.multiply(self.TF_left_blocked_previous[irBlockCount],
                                                       self.FDL_left[
                                                       (irBlockCount * (self.block_size + 1)):(
                                                           (irBlockCount + 1) * (self.block_size + 1))])

            self.resultRightFreqPrevious += np.multiply(self.TF_right_blocked_previous[irBlockCount],
                                                        self.FDL_right[
                                                        (irBlockCount * (self.block_size + 1)):(
                                                            (irBlockCount + 1) * (self.block_size + 1))])

    def process(self, block):
        """
        Main function

        :param block:
        :return: (outputLeft, outputRight)
        """
        # print("Convolver: process")

        # First: Fill buffer and FDLs with current block
        if not self.processStereo:
            # print('Convolver Mono Processing')
            self.fill_buffer_mono(block)
        else:
            # print('Convolver Stereo Processing')
            self.fill_buffer_stereo(block)

        # Second: Multiplikation with IR block und accumulation with previous data
        for irBlockCount in xrange(0, self.IR_blocks):
            # Always convolute current filter
            self.multiply_and_add(irBlockCount)

            # Also convolute old filter if interpolation needed
            if self.interpolate:
                self.multiply_and_add_previous(irBlockCount)

        # Third: Transformation back to time domain
        if self.interpolate:
            # fade over full block size
            # print('do block interpolation')
            self.outputLeft = np.multiply(self.resultLeftPreviousIFFTPlan(self.resultLeftFreqPrevious).real[
                                          self.block_size:self.block_size * 2], self.crossFadeOut) + \
                              np.multiply(self.resultLeftIFFTPlan(self.resultLeftFreq).real[
                                          self.block_size:self.block_size * 2], self.crossFadeIn)

            self.outputRight = np.multiply(self.resultRightPreviousIFFTPlan(self.resultRightFreqPrevious).real[
                                           self.block_size:self.block_size * 2], self.crossFadeOut) + \
                               np.multiply(self.resultRightIFFTPlan(self.resultRightFreq).real[
                                           self.block_size:self.block_size * 2], self.crossFadeIn)

        else:
            self.outputLeft = self.resultLeftIFFTPlan(self.resultLeftFreq).real[self.block_size:self.block_size * 2]
            self.outputRight = self.resultRightIFFTPlan(self.resultRightFreq).real[self.block_size:self.block_size * 2]

        self.processCounter += 1
        self.interpolate = False

        return self.outputLeft, self.outputRight

    def close(self):
        print("Convolver: close")
        # TODO: do something here?



# This file is part of the pyBinSim project.
#
# Copyright (c) 2017 A. Neidhardt, F. Klein, N. Knoop, T. Köllmer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import multiprocessing

import numpy as np
import soundfile as sf

from pybinsim.pose import Pose
from pybinsim.utility import total_size

nThreads = multiprocessing.cpu_count()


class FilterStorage(object):
    """ Class for storing all filters mentioned in the filter list """

    def __init__(self, irSize, block_size, filter_list_name):

        self.log = logging.getLogger("pybinsim.FilterStorage")
        self.log.info("FilterStorage: init")

        self.ir_size = irSize
        self.ir_blocks = irSize // block_size
        self.block_size = block_size
        self.default_filter = np.zeros((self.ir_size, 2), dtype='float32')

        self.filter_list_path = filter_list_name
        self.filter_list = open(self.filter_list_path, 'r')

        self.headphone_filter = None

        # format: [key,{filter}]
        self.filter_dict = {}

        # Start to load filters
        self.load_filters()

    def parse_filter_list(self):
        """
        Generator for filter list lines

        Lines are assumed to have a format like
        0 0 40 1 1 0 brirWav_APA/Ref_A01_1_040.wav

        The headphone filter starts with HPFILTER instead of the positions.

        Lines can be commented with a '#' as first character.

        :return: Iterator of (Pose, filter-path) tuples
        """

        for line in self.filter_list:

            # comment out lines in the list with a '#'
            if line.startswith('#'):
                continue

            line_content = line.split()
            filter_path = line_content[-1]

            if line.startswith('HPFILTER'):
                self.log.info("Loading headphone filter: {}".format(filter_path))
                self.headphone_filter = self.load_filter(filter_path)
                continue

            filter_value_list = tuple(line_content[0:-1])

            pose = Pose.from_filterValueList(filter_value_list)

            yield pose, filter_path

    def load_filters(self):
        """
        Load filters from files

        :return: None
        """

        self.log.info("Start loading filters...")

        for i, (pose, filter_path) in enumerate(self.parse_filter_list()):
            self.log.debug('Loading {}'.format(filter_path))

            filter = self.load_filter(filter_path)

            # create key and store in dict.
            key = pose.create_key()
            self.filter_dict.update({key: filter})

        self.log.info("Finished loading filters.")
        self.log.info("filter_dict size: {}MiB".format(total_size(self.filter_dict) // 1024 // 1024))

    def get_filter(self, pose):
        """
        Searches in the dict if key is available and return corresponding filter
        When no filter is found, defaultFilter is returned which results in silence

        :param pose
        :return: corresponding filter for pose
        """

        key = pose.create_key()

        if key in self.filter_dict:
            self.log.info('Filter found: key: {}'.format(key))
            return self.filter_dict.get(key)
        else:
            self.log.warning('Filter not found: key: {}'.format(key))
            return self.default_filter

    def close(self):
        self.log.info('FilterStorage: close()')
        # TODO: do something in here?

    def get_headphone_filter(self):
        if self.headphone_filter is None:
            raise RuntimeError("Headphone filter not loaded")

        return self.headphone_filter

    def load_filter(self, filter_path):

        current_filter, fs = sf.read(filter_path, dtype='float32')

        filter_size = np.shape(current_filter)

        # Fill filter with zeros if to short
        if filter_size[0] < self.ir_size:
            self.log.warning('Filter to short: Fill up with zeros')
            current_filter = np.concatenate((current_filter, np.zeros((self.ir_size - filter_size[0], 2))), 0)

        return current_filter



# This file is part of the pyBinSim project.
#
# Copyright (c) 2017 A. Neidhardt, F. Klein, N. Knoop, T. Köllmer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Module for interacting with sparkFun 9DOF Razor IMU
Product URL: https://www.sparkfun.com/products/10736
"""

from __future__ import print_function

import re

import serial


def get_intact_reading(sensor_reading, prefix='!ANG:'):
    """
    Return last intact sensor data (i.e., the newest) in a list of readings.
    :param sensor_reading: List of sender readings, e.g. from read_all() split into lines.
    :param prefix: Prefix of sensor reading, e.g. "!ANG"
    :return: CSV string without prefix. None if parsing failed.
    """
    digit_regex = r"(([-]?\d*\.\d+)|[-]?\d+)?"
    reading_regex = r"{}{},{},{}".format(prefix, digit_regex, digit_regex, digit_regex)

    for item in reversed(sensor_reading):
        item = item.strip()
        if re.match(reading_regex, item):
            return item[len(prefix):]
    return None


def get_float_values(line):
    """
    Parse csv string with float values to list of floats.
    :param line:  csv string, e.g., ".4,2,-3.4"
    :return: List of floats. Empty list if parsing failed.
    """
    result_list = []
    yrp = line.split(",")

    for x in yrp:
        if x.strip() == "":
            result_list += [0]
        else:
            try:
                result_list += [float(x)]
            except:
                print("Could not parse line: {}", line)
                return []

    return result_list


def parse_sensor_reading(sensor_reading):
    """
    Parses sensor reading and returns list of floats.
    :param sensor_reading: List of sender readings, e.g. from read_all() split into lines.
    :return: List of floats with sensor values. Empty list if parsing failed.
    """
    if len(sensor_reading) == 0:
        return []

    line = get_intact_reading(sensor_reading)

    if not line:
        return []

    result_list = get_float_values(line)
    return result_list


class Spark9dof(object):
    """
    Class allows access to Yaw, Pitch, Roll Data of spark fun's 9DoF Board.
    """

    def __init__(self, com_port='COM4', baudrate=57600):
        """
        Initiializes Serial connection to 9DOF board
        :param com_port: COM Port, defaults to 'COM4'.
        :param baudrate: baudrate, defaults to 57600
        """

        self.com_port = com_port
        self.baudrate = baudrate

        try:
            self.ser = serial.Serial(self.com_port, self.baudrate, timeout=1)
        except serial.SerialException as e:
            raise RuntimeError(e)

    def get_sensor_data(self):
        """
        Return parsed sensor_reading.

        Returns the most recent value in all sensor readings that have been cached
        between two calls to get_sensor_data. Beware that too short polling times might
        result in no values returned, as the buffer does not (yet) contain valid sensor
        readings, but only fragments.
        :return: Sensor reading as list of float.
        """
        sensor_reading = self.ser.read_all()
        reading_list = sensor_reading.decode().split("\r\n")

        return parse_sensor_reading(reading_list)



from collections import namedtuple


class Orientation(namedtuple('Orientation', ['yaw', 'pitch', 'roll'])):
    pass


class Position(namedtuple('Position', ['x', 'y', 'z'])):
    pass


class Pose:
    def __init__(self, orientation, position):
        self.orientation = orientation
        self.position = position

    def create_key(self):
        value_list = list(self.orientation) + list(self.position)

        return ','.join([str(x) for x in value_list])

    @staticmethod
    def from_filterValueList(filter_value_list):
        if len(filter_value_list) != 6:
            print(filter_value_list)
            assert False

        orientation = Orientation(filter_value_list[0], filter_value_list[1], filter_value_list[2])
        position = Position(filter_value_list[3], filter_value_list[4], filter_value_list[5])

        return Pose(orientation, position)



# This file is part of the pyBinSim project.
#
# Copyright (c) 2017 A. Neidhardt, F. Klein, N. Knoop, T. Köllmer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""" Module contains main loop and configuration of pyBinSim """
import logging
import time

import numpy as np
import pyaudio

from pybinsim.convolver import ConvolverFFTW
from pybinsim.filterstorage import FilterStorage
from pybinsim.osc_receiver import OscReceiver
from pybinsim.pose import Pose
from pybinsim.soundhandler import SoundHandler


class BinSimConfig(object):
    def __init__(self):

        self.log = logging.getLogger("pybinsim.BinSimConfig")

        # Default Configuration
        self.configurationDict = {'soundfile': '',
                                  'blockSize': 256,
                                  'filterSize': 16384,
                                  'filterList': 'brirs/filter_list_kemar5.txt',
                                  'enableCrossfading': 'False',
                                  'useHeadphoneFilter': 'False',
                                  'loudnessFactor': float(1),
                                  'maxChannels': 8,
                                  'samplingRate': 44100,
                                  'loopSound': 'True'}

    def read_from_file(self, filepath):
        config = open(filepath, 'r')

        for line in config:
            line_content = str.split(line)
            key = line_content[0]
            if key in self.configurationDict:
                self.configurationDict[key] = type(self.configurationDict[key])(line_content[1])
            else:
                self.log.warning('Entry ' + key + ' is unknown')

    def get(self, setting):
        return self.configurationDict[setting]


class BinSim(object):
    """
    Main pyBinSim program logic
    """

    def __init__(self, config_file):

        self.log = logging.getLogger("pybinsim.BinSim")
        self.log.info("BinSim: init")

        # Read Configuration File
        self.config = BinSimConfig()
        self.config.read_from_file(config_file)

        self.current_config = self.config
        self.nChannels = self.current_config.get('maxChannels')
        self.sampleRate = self.current_config.get('samplingRate')
        self.blockSize = self.current_config.get('blockSize')

        self.result = None
        self.block = None
        self.stream = None

        self.convolverHP, self.convolvers, self.filterStorage, self.oscReceiver, self.soundHandler = self.initialize_pybinsim()

        self.p = pyaudio.PyAudio()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__cleanup()

    def stream_start(self):
        self.log.info("BinSim: stream_start")
        self.stream = self.p.open(format=pyaudio.paFloat32, channels=2,
                                  rate=self.sampleRate, output=True,
                                  frames_per_buffer=self.blockSize,
                                  stream_callback=audio_callback(self))
        self.stream.start_stream()

        while self.stream.is_active():
            time.sleep(1)

    def initialize_pybinsim(self):
        self.result = np.empty([self.config.get('blockSize'), 2], np.dtype(np.float32))
        self.block = np.empty([self.config.get('maxChannels'), self.config.get('blockSize')], np.dtype(np.float32))

        # Create FilterStorage
        filterStorage = FilterStorage(self.config.get('filterSize'), self.config.get('blockSize'),
                                      self.config.get('filterList'))

        # Start an oscReceiver
        oscReceiver = OscReceiver()
        oscReceiver.start_listening()
        time.sleep(1)

        # Create SoundHandler
        soundHandler = SoundHandler(self.config.get('blockSize'), self.config.get('maxChannels'),
                                    self.config.get('samplingRate'), self.config.get('loopSound'))

        soundfile_list = self.config.get('soundfile')
        soundHandler.request_new_sound_file(soundfile_list)

        # Create N convolvers depending on the number of wav channels
        self.log.info('Number of Channels: ' + str(self.config.get('maxChannels')))
        convolvers = [None] * self.config.get('maxChannels')
        for n in range(self.config.get('maxChannels')):
            convolvers[n] = ConvolverFFTW(self.config.get('filterSize'), self.config.get('blockSize'), False)

        # HP Equalization convolver
        convolverHP = None
        if self.config.get('useHeadphoneFilter') == 'True':
            convolverHP = ConvolverFFTW(self.config.get('filterSize'), self.config.get('blockSize'), True)
            hpfilter = filterStorage.get_headphone_filter()
            convolverHP.setIR(hpfilter, False)

        return convolverHP, convolvers, filterStorage, oscReceiver, soundHandler

    def close(self):
        self.log.info("BinSim: close")
        self.stream_close()
        self.p.terminate()

    def stream_close(self):
        self.log.info("BinSim: stream_close")
        self.stream.stop_stream()
        self.stream.close()

    def __cleanup(self):
        # Close everything when BinSim is finished
        self.filterStorage.close()
        self.close()

        self.oscReceiver.close()

        for n in range(self.config.get('maxChannels')):
            self.convolvers[n].close()

        if self.config.get('useHeadphoneFilter') == 'True':
            if self.convolverHP:
                self.convolverHP.close()


def audio_callback(binsim):
    """ Wrapper for callback to hand over custom data """
    assert isinstance(binsim, BinSim)

    # The pyAudio Callback
    def callback(in_data, frame_count, time_info, status):
        # print("pyAudio callback")

        current_soundfile_list = binsim.oscReceiver.get_sound_file_list()
        if current_soundfile_list:
            binsim.soundHandler.request_new_sound_file(current_soundfile_list)

        # Get sound block. At least one convolver should exist
        binsim.block[:binsim.soundHandler.get_sound_channels(), :] = binsim.soundHandler.buffer_read()

        # Update Filters and run each convolver with the current block
        for n in range(binsim.soundHandler.get_sound_channels()):

            # Get new Filter
            if binsim.oscReceiver.is_filter_update_necessary(n):
                # print('Updating Filter')
                filterValueList = binsim.oscReceiver.get_current_values(n)
                filter = binsim.filterStorage.get_filter(Pose.from_filterValueList(filterValueList))
                binsim.convolvers[n].setIR(filter, callback.config.get('enableCrossfading'))

            left, right = binsim.convolvers[n].process(binsim.block[n, :])

            # Sum results from all convolvers
            if n == 0:
                binsim.result[:, 0] = left
                binsim.result[:, 1] = right
            else:
                binsim.result[:, 0] += left
                binsim.result[:, 1] += right

        # Finally apply Headphone Filter
        if callback.config.get('useHeadphoneFilter') == 'True':
            binsim.result[:, 0], binsim.result[:, 1] = binsim.convolverHP.process(binsim.result)

        # Scale data
        binsim.result *= 1 / float((callback.config.get('maxChannels') + 1) * 2)
        binsim.result *= callback.config.get('loudnessFactor')

        # When the last block is small than the blockSize, this is probably the end of the file.
        # Call pyaudio to stop after this frame
        if binsim.block.size < callback.config.get('blockSize'):
            pyaudio.paContinue = 1

        return (binsim.result[:frame_count].tostring(), pyaudio.paContinue)

    callback.config = binsim.config

    return callback



# coding=utf-8
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

VERSION = "1.1.1"

class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(["--junitxml", "junit_results.xml"])
        sys.exit(errno)


setup(
    name='pybinsim',
    version=VERSION,
    license='MIT',
    author='Annika Neidhardt, Florian Klein, Thomas Koellmer',
    author_email='thomas.koellmer@tu-ilmenau.de',
    url='https://github.com/pyBinSim/pyBinSim',
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
    install_requires=[
        "future >= 0.16.0",
        "numpy >= 1.12.1",
        "ovr >= 1.10.101",
        "pyaudio >= 0.2.10",
        "pyfftw >= 0.10.4",
        "pyserial >= 3.2.1",
        "pytest >= 3.0.5",
        "python-osc >= 1.6.3",
        "six >= 1.10.0",
        "Soundfile >= 0.9.0",
    ],

    description='Real-time dynamic binaural synthesis with head tracking.',
    long_description=open('README.rst').read(),
    packages=['pybinsim'],
    include_package_data=True,
    platforms='any',
    data_files=[],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Multimedia :: Sound/Audio',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License'
    ],

)



# This file is part of the pyBinSim project.
#
# Copyright (c) 2017 A. Neidhardt, F. Klein, N. Knoop, T. Köllmer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import threading

from pythonosc import dispatcher
from pythonosc import osc_server


class OscReceiver(object):
    """
    Class for receiving OSC Messages to control pyBinSim
    """

    def __init__(self):

        self.log = logging.getLogger("pybinsim.OscReceiver")
        self.log.info("oscReceiver: init")

        # Basic settings
        self.ip = '127.0.0.1'
        self.port = 10000
        self.maxChannels = 100

        # Default values; Stores filter keys for all channles/convolvers
        self.filters_updated = [True] * self.maxChannels

        self.defaultValue = (0, 0, 0, 0, 0, 0)
        self.valueList = [self.defaultValue] * self.maxChannels
        # self.valueList = [()] * self.maxChannels
        self.soundFileList = ''
        self.soundFileNew = False

        osc_dispatcher = dispatcher.Dispatcher()
        osc_dispatcher.map("/pyBinSim", self.handle_filter_input)
        osc_dispatcher.map("/pyBinSimFile", self.handle_file_input)

        self.server = osc_server.ThreadingOSCUDPServer(
            (self.ip, self.port), osc_dispatcher)

    def handle_filter_input(self, identifier, channel, *args):
        """
        Handler for tracking information

        :param identifier:
        :param channel:
        :param args:
        :return:
        """

        assert identifier == "/pyBinSim"
        # assert all(isinstance(x, int) for x in args) == True

        # Extend value list to support older scripts
        # if (len(args)<6):
        #    args=(args+(0,)*6)[:6]
        #    print("filter value list incomplete")

        self.log.info("Channel: {}".format(str(channel)))
        self.log.info("Args: {}".format(str(args)))

        current_channel = channel

        if args != self.valueList[current_channel]:
            self.log.info("new filter")
            self.filters_updated[current_channel] = True
            self.valueList[current_channel] = tuple(args)
        else:
            self.log.info("same filter as before")

    def handle_file_input(self, identifier, soundpath):
        """ Handler for playlist control"""

        assert identifier == "/pyBinSimFile"
        # assert type(soundpath) == 'str'

        self.log.info("soundPath: {}".format(soundpath))
        self.soundFileList = soundpath

    def start_listening(self):
        """Start osc receiver in background Thread"""

        self.log.info("Serving on {}".format(self.server.server_address))

        osc_thread = threading.Thread(target=self.server.serve_forever)
        osc_thread.daemon = True
        osc_thread.start()

    def is_filter_update_necessary(self, channel):
        """ Check if there is a new filter for channel """
        return self.filters_updated[channel]

    def get_current_values(self, channel):
        """ Return key for filter """
        self.filters_updated[channel] = False
        return self.valueList[channel]

    def get_sound_file_list(self):
        ret_list = self.soundFileList
        self.soundFileList = ''
        return ret_list

    def close(self):
        """
        Close the osc receiver

        :return: None
        """
        self.log.info('oscReiver: close()')
        self.server.shutdown()



# Taken from https://github.com/mgeier/python-audio/blob/master/audio-files/utility.py


"""Helper functions for working with audio files in NumPy."""
import contextlib

from sys import getsizeof, stderr
from itertools import chain
from collections import deque

import numpy as np


def pcm2float(sig, dtype='float64'):
    """Convert PCM signal to floating point with a range from -1 to 1.
    Use dtype='float32' for single precision.
    Parameters
    ----------
    sig : array_like
        Input array, must have integral type.
    dtype : data type, optional
        Desired (floating point) data type.
    Returns
    -------
    numpy.ndarray
        Normalized floating point data.
    See Also
    --------
    float2pcm, dtype
    """
    sig = np.asarray(sig)
    if sig.dtype.kind not in 'iu':
        raise TypeError("'sig' must be an array of integers")
    dtype = np.dtype(dtype)
    if dtype.kind != 'f':
        raise TypeError("'dtype' must be a floating point type")

    i = np.iinfo(sig.dtype)
    abs_max = 2 ** (i.bits - 1)
    offset = i.min + abs_max
    return (sig.astype(dtype) - offset) / abs_max


def float2pcm(sig, dtype='int16'):
    """Convert floating point signal with a range from -1 to 1 to PCM.
    Any signal values outside the interval [-1.0, 1.0) are clipped.
    No dithering is used.
    Note that there are different possibilities for scaling floating
    point numbers to PCM numbers, this function implements just one of
    them.  For an overview of alternatives see
    http://blog.bjornroche.com/2009/12/int-float-int-its-jungle-out-there.html
    Parameters
    ----------
    sig : array_like
        Input array, must have floating point type.
    dtype : data type, optional
        Desired (integer) data type.
    Returns
    -------
    numpy.ndarray
        Integer data, scaled and clipped to the range of the given
        *dtype*.
    See Also
    --------
    pcm2float, dtype
    """
    sig = np.asarray(sig)
    if sig.dtype.kind != 'f':
        raise TypeError("'sig' must be a float array")
    dtype = np.dtype(dtype)
    if dtype.kind not in 'iu':
        raise TypeError("'dtype' must be an integer type")

    i = np.iinfo(dtype)
    abs_max = 2 ** (i.bits - 1)
    offset = i.min + abs_max
    return (sig * abs_max + offset).clip(i.min, i.max).astype(dtype)


def pcm24to32(data, channels=1, normalize=True):
    """Convert 24-bit PCM data to 32-bit.
    Parameters
    ----------
    data : buffer
        A buffer object where each group of 3 bytes represents one
        little-endian 24-bit value.
    channels : int, optional
        Number of channels, by default 1.
    normalize : bool, optional
        If ``True`` (the default) the additional zero-byte is added as
        least significant byte, effectively multiplying each value by
        256, which leads to the maximum 24-bit value being mapped to the
        maximum 32-bit value.  If ``False``, the zero-byte is added as
        most significant byte and the values are not changed.
    Returns
    -------
    numpy.ndarray
        The content of *data* converted to an *int32* array, where each
        value was padded with zero-bits in the least significant byte
        (``normalize=True``) or in the most significant byte
        (``normalize=False``).
    """
    if len(data) % 3 != 0:
        raise ValueError('Size of data must be a multiple of 3 bytes')

    out = np.zeros(len(data) // 3, dtype='<i4')
    out.shape = -1, channels
    temp = out.view('uint8').reshape(-1, 4)
    if normalize:
        # write to last 3 columns, leave LSB at zero
        columns = slice(1, None)
    else:
        # write to first 3 columns, leave MSB at zero
        columns = slice(None, -1)
    temp[:, columns] = np.frombuffer(data, dtype='uint8').reshape(-1, 3)
    return out


@contextlib.contextmanager
def printoptions(*args, **kwargs):
    """Context manager for temporarily setting NumPy print options.
    See http://stackoverflow.com/a/2891805/500098
    """
    original = np.get_printoptions()
    try:
        np.set_printoptions(*args, **kwargs)
        yield
    finally:
        np.set_printoptions(**original)



# taken from https://code.activestate.com/recipes/577504/ as recommended by
# https://docs.python.org/3.5/library/sys.html#sys.getsizeof
def total_size(o, handlers={}, verbose=False):
    """ Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {tuple: iter,
                    list: iter,
                    deque: iter,
                    dict: dict_handler,
                    set: iter,
                    frozenset: iter,
                   }
    all_handlers.update(handlers)     # user handlers take precedence
    seen = set()                      # track which object id's have already been seen
    default_size = getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


# This file is part of the pyBinSim project.
#
# Copyright (c) 2017 A. Neidhardt, F. Klein, N. Knoop, T. Köllmer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import threading
import time

import numpy as np
import soundfile as sf


class SoundHandler(object):
    """ Class to read audio from files and serve it to pyBinSim """

    def __init__(self, block_size, n_channels, fs, loopSound):

        self.log = logging.getLogger("pybinsim.SoundHandler")

        self.fs = fs
        self.n_channels = n_channels
        self.chunk_size = block_size
        self.bufferSize = block_size * 2
        self.buffer = np.zeros([self.n_channels, self.bufferSize])
        self.sound = np.empty((0, 0))
        self.sound_file = np.empty((0, 0))
        self.frame_count = 0
        self.active_channels = 0
        self.soundPath = ''
        self.new_sound_file_request = False
        self.new_sound_file_loaded = False
        self.loopSound = loopSound
        self.currentSoundFile = 1
        self.soundFileList = []

        self._run_file_reader()

    def buffer_add_silence(self):
        self.buffer[:self.active_channels, :-self.chunk_size] = self.buffer[:self.active_channels, self.chunk_size:]
        self.buffer[:self.active_channels, -self.chunk_size:] = np.zeros([self.active_channels, self.chunk_size])

    def buffer_add_sound(self):
        if (self.frame_count + 1) * self.chunk_size < self.sound.shape[1]:
            self.buffer[:self.active_channels, :-self.chunk_size] = self.buffer[:self.active_channels, self.chunk_size:]
            self.buffer[:self.active_channels, -self.chunk_size:] = self.sound[
                                                                    :self.active_channels,
                                                                    self.frame_count * self.chunk_size: (
                                                                                                        self.frame_count + 1) * self.chunk_size
                                                                    ]
            self.frame_count += 1
        elif self.currentSoundFile < len(self.soundFileList) and not self.new_sound_file_request:
            self.request_next_sound_file()
        elif self.loopSound == 'True' and not self.new_sound_file_request:
            self.currentSoundFile = 0
            self.request_next_sound_file()
            self.frame_count = 0
        else:
            self.buffer_add_silence()

    def buffer_flush(self):
        self.buffer = np.zeros([self.n_channels, self.bufferSize])

    def buffer_read(self):
        if self.new_sound_file_loaded:
            self.buffer_flush()
            self.sound = self.sound_file
            self.frame_count = 0
            self.new_sound_file_loaded = False

        buffer_content = self.buffer[:self.active_channels, :-self.chunk_size]
        self.buffer_add_sound()
        return buffer_content

    def _run_file_reader(self):
        file_read_thread = threading.Thread(target=self.read_sound_file)
        file_read_thread.start()

    def read_sound_file(self):

        while True:
            if self.new_sound_file_request:
                self.log.info('Loading new sound file')
                audio_file_data, fs = sf.read(self.soundPath, dtype='float32', )
                assert fs == self.fs

                self.log.debug("audio_file_data: {} MB".format(audio_file_data.nbytes // 1024 // 1024))
                self.sound_file = np.asmatrix(audio_file_data)

                # free data
                audio_file_data = None

                if self.sound_file.shape[0] > self.sound_file.shape[1]:
                    self.sound_file = self.sound_file.transpose()

                self.active_channels = self.sound_file.shape[0]

                if self.sound_file.shape[1] % self.chunk_size != 0:
                    length_diff = self.chunk_size - (self.sound_file.shape[1] % self.chunk_size)
                    zeros = np.zeros((self.sound_file.shape[0], length_diff), dtype=np.float32)

                    self.log.debug("Zeros size: {} Byte".format(zeros.nbytes))
                    self.log.debug("Zeros shape: {} ({})".format(zeros.shape, zeros.dtype))
                    self.log.debug("Soundfile size: {} MiB".format(self.sound_file.nbytes // 1024 // 1024))
                    self.log.debug("Soundfile shape: {} ({})".format(self.sound_file.shape, self.sound_file.dtype))
                    self.sound_file = np.concatenate(
                        (self.sound_file, zeros),
                        1
                    )
                    self.log.debug("Soundfile size after concat: {} MiB".format(self.sound_file.nbytes // 1024 // 1024))
                    self.log.debug(
                        "Soundfile shape after concat: {} ({})".format(self.sound_file.shape, self.sound_file.dtype))
                    self.log.info('Loaded new sound file\n')
                self.new_sound_file_request = False
                self.new_sound_file_loaded = True
            time.sleep(0.05)

    def request_new_sound_file(self, sound_file_list):

        sound_file_list = str.split(sound_file_list, '#')
        self.soundPath = sound_file_list[0]
        self.soundFileList = sound_file_list
        self.currentSoundFile = 1
        self.new_sound_file_request = True

    def request_next_sound_file(self):

        self.currentSoundFile += 1
        self.soundPath = self.soundFileList[self.currentSoundFile - 1]
        self.new_sound_file_request = True

    def get_sound_channels(self):
        return self.active_channels



import logging

__version__ = "1.1.1"

from pybinsim.application import BinSim


def init_logging(loglevel):
    console_handler = logging.StreamHandler()
    console_handler.setLevel(loglevel)

    formatter = logging.Formatter('%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)

    logger = logging.getLogger("pybinsim")
    logger.addHandler(console_handler)
    logger.setLevel(loglevel)

    return logger


logger = init_logging(logging.INFO)
